# Standard imports
from threading import Thread as T


class Thread(T):
    """ A thread that can return a value once it's done. """

    def __init__(self, group = None, target = None, name = None, args = (), kwargs = {}, Verbose = None):
        T.__init__(self, group, target, name, args, kwargs)
        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)

    def join(self, *args):
        T.join(self, *args)
        return self._return


class ThreadPool():
    """ A pool of threads that can return a list of return values from threads once it's done """

    def __init__(self, confs: dict) -> None:
        """
        Constructor

        Parameters:
            confs: A dict of tuples containing the following keys:
                - key: The name of the thread
                - value: A tuple containing the following:
                    - target: The function/callable to be ran in the thread
                    - args: A tuple of arguments to be passed to the function
        """
        self.threads = []

        # Create threads
        for name, (target, args) in confs.items():
            if args is None: args = ()
            self.threads.append(Thread(target = target, args = args, name = name))


    def start(self) -> None:
        """ Start all threads in parallel """
        for thread in self.threads:
            thread.start()


    def join(self) -> dict:
        """ Wait for all threads to finish and return a dictionary of name:<return value> """
        return {thread.name: thread.join() for thread in self.threads}


# Standard imports
from .threads import Thread
from requests import Response, Session
from requests.adapters import HTTPAdapter, Retry


class BaseRequests():

    def __init__(self, timeout: int = None, retries: int = 3, backoff_factor: float = 0.1, status_forcelist: list = [500, 502, 503, 504]) -> None:
        """
            Initialize a Base Requests instance with a Backoff-Strategy

            Parameters:
                timeout (int): The number of seconds to wait before throwing a TimeOutExceotion (default is None, which means it'll wait until the connection is closed). Be careful with this, as it can exhaust retries.
                retries (int): The number of times a request is rest with a Backoff-Strategy before giving up and throwign an Exception
                backoff_factor (float): Increase the amout of wait time between each failed requests
                status_forcelist (list): List of error messages that are forced a retry if received
        """

        self.timeout = timeout
        self.session = Session()
        retries = Retry(
            total = retries,
            backoff_factor = backoff_factor,
            status_forcelist = status_forcelist
        )

        # Mount session
        self.session.mount('http://', HTTPAdapter(max_retries = retries))
        self.session.mount('https://', HTTPAdapter(max_retries = retries))


    def get(self, url: str, params: dict = {}, headers: dict = None, threaded: bool = False) -> Response:
        """ GET request """
        thread = Thread(target = self.session.get, kwargs = {"url": url, "params": params, "headers": headers, "timeout": self.timeout})
        thread.start()
        if not threaded: return thread.join()


    def post(self, url: str, json: dict = {}, params: dict = {}, headers: dict = None, threaded: bool = False) -> Response:
        """ POST request """
        thread = Thread(target = self.session.post, kwargs = {"url": url, "json": json, "params": params, "headers": headers, "timeout": self.timeout})
        thread.start()
        if not threaded: return thread.join()


    def delete(self, url: str, json: dict = {}, params: dict = {}, headers: dict = None, threaded: bool = False) -> Response:
        """ DELETE request """
        thread = Thread(target = self.session.delete, kwargs = {"url": url, "json": json, "params": params, "headers": headers, "timeout": self.timeout})
        thread.start()
        if not threaded: return thread.join()


    def put(self, url: str, json: dict = {}, params: dict = {}, headers: dict = None, threaded: bool = False) -> Response:
        """ PUT request """
        thread = Thread(target = self.session.put, kwargs = {"url": url, "json": json, "params": params, "headers": headers, "timeout": self.timeout})
        thread.start()
        if not threaded: return thread.join()


# Standard imports
import yaml

def read_config(project_name: str, path: str = "./config.yml") -> dict:
    """ Reads the config file and returns the configurations as a dictionary """
    with open(path, "r") as file:
        return yaml.safe_load(file)[project_name]